/*
* ple.c
* polygonal line editor
* Luiz Henrique de Figueiredo
* 28 May 93
*/

#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include "defs.h"
#include "laux.h"
#include "geom.h"
#include "gp.h"
#include "cb.h"

#define	TOL	1e-2

typedef struct point	Point;

struct point {
 real	x,y; 
 Point*	next;
 Point*	prev;
};

void	redraw		(int clear);
void	delpoints	(void);
void	showpolygon	(void);
void	showspline	(void);
void	showpoints	(void);
void	addpoint	(real x, real y);
void	movepoint	(real x, real y);
void	delpoint	(real x, real y);
void	startmove	(real x, real y);
void	endmove		(real x, real y);
void	showchange	(Point* p, int c);
void	showpoint	(Point* p);
void	showside	(Point* p, Point *q);
Point*	findpoint	(real x, real y);
void    read_curve       (char *fname);

Callback 
 do_test,
 do_clear,
 do_polygon,
 do_quit,
 do_redraw,
 do_addpoint,
 do_startmove,
 do_endmove,
 do_delpoint,
 do_movepoint;

#define	X(p)		((p)->x)
#define	Y(p)		((p)->y)
#define	new(t)		((t*)emalloc(sizeof(t)))
#define	next(p)		((p)->next)
#define	prev(p)		((p)->prev)

static Point*		firstpoint=NULL;
static Point*		lastpoint=NULL;
static int		showingpolygon=1;
static int		showingpoints=1;
static real             xmin = 0, xmax = 1, ymin = 0, ymax = 1;

int main(int argc, char* argv[])
{
 if (argc == 2)
   read_curve(argv[1]);
 gpopen("polygonal line editor");
 gpwindow(xmin,xmax, ymin,ymax);
 gpmark(0,"B");				/* filled box mark */
 redraw(1);
 gpregister("kz",do_test,0);
 gpregister("kc",do_clear,0);
 gpregister("kp",do_polygon,0);
 gpregister("kq",do_quit,0);
 gpregister("kr",do_redraw,0);
 gpregister("k\f",do_redraw,0);
 gpregister("b1+",do_addpoint,0);
 gpregister("b2+",do_startmove,0);
 gpregister("b2-",do_endmove,0);
 gpregister("b3-",do_delpoint,0);
 gpregister("m2+",do_movepoint,0);
 gpmainloop();
 gpclose(0);
}

void redraw(int clear)
{
 if (clear)
   gpclear(0);
 if (showingpolygon)
   showpolygon();
 showpoints();
}

void delpoints(void)
{
 firstpoint=lastpoint=NULL;			/* lazy! */
}

Point* moving=NULL;

void addpoint(real x, real y)
{
 Point* p=new(Point);
 X(p)=x;
 Y(p)=y;
 next(p)=NULL;
 if (showingpoints) showpoint(p);
 if (firstpoint==NULL) {
  prev(p)=NULL;
  firstpoint=p;
 } else {
  prev(p)=lastpoint;   next(lastpoint)=p;
  if (showingpolygon) showside(lastpoint,p);
 }
 lastpoint=p;
}

void delpoint(real x, real y)
{
 Point* p=findpoint(x,y);
 if (p!=NULL) {
  if (prev(p)==NULL) firstpoint=next(p); else next(prev(p))=next(p);
  if (next(p)==NULL) lastpoint=prev(p);  else prev(next(p))=prev(p);
  redraw(1);
 }
}

void startmove(real x, real y)
{
 moving=findpoint(x,y);
 if (moving!=NULL) {
  x=X(moving); y=Y(moving);
  gpcolor(0);    gpplot(x,y); gpcolor(1);
  gpmark(0,"b"); gpplot(x,y);
 }
}

void movepoint(real x, real y)
{
 if (moving!=NULL) {
  showchange(moving,0);
  X(moving)=x; Y(moving)=y;
  showchange(moving,1);
 }
 else startmove(x,y);
}

void endmove(real x, real y)
{
 if (moving!=NULL) {
  gpmark(0,"B");
  redraw(0);
  moving=NULL;
 }
}

Point* findpoint(real x, real y)
{
 Point* p=firstpoint;
 for (p=firstpoint; p!=NULL; p=next(p))
  if ((fabs(X(p)-x)+fabs(Y(p)-y))<TOL) break;
 return p;
}

void showpoints(void)
{
 Point* p;
 for (p=firstpoint; p!=NULL; p=next(p))
  showpoint(p);
}

void showpolygon(void)
{
 Point* p;
 for (p=firstpoint; p!=NULL; p=next(p))
  showside(p,next(p));
}

void showpoint(Point* p)
{
 gpplot(X(p),Y(p));
}

void showside(Point* p, Point *q)
{
 if (p!=NULL && q!=NULL) gpline(X(p),Y(p),X(q),Y(q));
}

void showchange(Point* p, int c)
{
 gpcolor(c);
 showpoint(p);
 if (showingpolygon) {
  showside(prev(p),p);
  showside(p,next(p));
 }
}

int do_test(char* e, real x, real y, void* q)
{
  Point* p;
  Vector3 g[4], v1, v2, v3, n1a, n2a, n1b, n2b, n1;
  Real d1, d2, s1a, s1b, s2a, s2b, eps = REL_EPS;
  int i;

  for (i=0,p=firstpoint; p!=NULL; i++,p=next(p))  {
    if (i == 4)
      break;
    g[i] = v3_make(X(p), Y(p), 0);
  }
  if (i < 4)
    return 0;

  v1 = v3_sub(g[1], g[3]);
  v2 = v3_sub(g[1], g[0]);
  v3 = v3_sub(g[1], g[2]);

  n1a = v3_cross(v2, v1); s1a = v3_sqrnorm(n1a);
  n2a = v3_cross(v2, v3); s2a = v3_sqrnorm(n2a);

  v1 = v3_sub(g[3], g[1]);
  v2 = v3_sub(g[3], g[2]);
  v3 = v3_sub(g[3], g[0]);

  n1b = v3_cross(v2, v1); s1b = v3_sqrnorm(n1b);
  n2b = v3_cross(v2, v3); s2b = v3_sqrnorm(n2b);

  if (s1a > eps)
    if (s1b > eps)
      n1 = v3_unit(v3_add(n1a, n1b));
    else
      n1 = v3_unit(n1a);
  else
    if (s1b > eps)
      n1 = v3_unit(n1b);
    else
      n1 = v3_make(0,0,0);
  
  d1 = v3_dot(n1, n2a);
  if (d1 < 0)
    s2a *= -1;

  d2 = v3_dot(n1, n2b);
  if (d2 < 0)
    s2b *= -1;

  printf("\nRIGHT (%g %g),  dot %g\n",s1a, s1b, d1);
  printf("LEFT (%g %g),  dot %g\n", s2a, s2b, d2);

  printf("A %g\n", (MIN(s1a, s1b) / MAX(s1a, s1b)));
  printf("B %g\n", (MIN(s2a, s2b) / MAX(s2a, s2b)));

  return 0;
}

int do_clear(char* e, real x, real y, void* p)
{
 delpoints();
 redraw(1);
 return 0;
}

int do_polygon(char* e, real x, real y, void* p)
{
 showingpolygon=!showingpolygon;
 redraw(1);
 return 0;
}

int do_quit(char* e, real x, real y, void* p)
{
 return 1;
}

int do_redraw(char* e, real x, real y, void* p)
{
 redraw(1);
 return 0;
}

int do_addpoint(char* e, real x, real y, void* p)
{
 addpoint(x,y);
 return 0;
}

int do_startmove(char* e, real x, real y, void* p)
{
 startmove(x,y);
 return 0;
}

int do_endmove(char* e, real x, real y, void* p)
{
 endmove(x,y);
 return 0;
}

int do_delpoint(char* e, real x, real y, void* p)
{
 delpoint(x,y);
 return 0;
}

int do_movepoint(char* e, real x, real y, void* p)
{
 movepoint(x,y);
 return 0;
}

void read_curve(char *fname)
{
  real x, y, z, m; int first = 1;
  FILE* fp = fopen(fname, "r");

  if (fp == NULL) return;
  showingpoints = showingpolygon = 0;
  while (fscanf(fp, "%lf %lf %lf\n" ,&x,&y,&z) != EOF) {
    addpoint(x, y);
    if (first) {
      xmin = xmax = x;
      ymin = ymax = y;
      first = 0;
    }
    if (x < xmin) xmin = x;
    if (y < ymin) ymin = y;
    if (x > xmax) xmax = x;
    if (y > ymax) ymax = y;
  }
  m = MAX((xmax - xmin), (ymax - ymin));
  xmin -= m * 0.1;
  xmax += m * 0.1;
  ymin -= m * 0.1;
  ymax += m * 0.1;
  showingpoints = showingpolygon = 1;
}

